import argparse
import boto3
import datetime
import json
import random
import string
import time
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from itertools import zip_longest
from retrying import retry

# global variables
_rule_set_name = "NONE"
_waiting_players = {}
_matched_sessions_queue = deque()
_start_time = int(time.time())

# constant variables 
_parser = argparse.ArgumentParser(description='Generate random players, start matchmaking and get results')
_parser.add_argument('--config-name', default="BuildersFlashSampleConfig") 
_parser.add_argument('--max-list-length',type=int, default=3)
_parser.add_argument('--request-interval', type=float, default=0.4)
_args = _parser.parse_args() 
CONFIG_NAME = _args.config_name
MAX_MATCHED_LIST_LENGTH = _args.max_list_length
REQUEST_INTERVAL = _args.request_interval


print (REQUEST_INTERVAL)

def elapsed_time():
    return int(time.time() - _start_time)


# Add matching player into temporary list
@retry(stop_max_attempt_number=3)
def add_matching_players_list(player):
    _waiting_players[player["PlayerId"]] = player["PlayerAttributes"]
    _waiting_players[player["PlayerId"]]["StartTime"] = {"N": elapsed_time()}


@retry(stop_max_attempt_number=3)
def enqueue_matched_players(match):
    _matched_sessions_queue.append(match)


@retry(stop_max_attempt_number=3)
def clear_matched_queue():
    for i in range(len(_matched_sessions_queue) - MAX_MATCHED_LIST_LENGTH):
        _matched_sessions_queue.popleft()


# Get player attributes from FlexMatch RuleSet 
def get_player_attr_definition():
    global _rule_set_name
    gamelift = boto3.client("gamelift")
    response = gamelift.describe_matchmaking_configurations(Names=[CONFIG_NAME])
    _rule_set_name = response["Configurations"][0]["RuleSetName"]
    response = gamelift.describe_matchmaking_rule_sets(Names=[_rule_set_name])
    rulesetjson = json.loads(response["RuleSets"][0]["RuleSetBody"])
    return rulesetjson["playerAttributes"]


@retry(stop_max_attempt_number=3)
def start_matchmaking(players_attrs_def, target=None):
    gamelift = boto3.client("gamelift")
    Player = {
        "PlayerId": "".join(random.choices(string.ascii_lowercase, k=6)),
        "PlayerAttributes": {},
    }

    for attr in players_attrs_def:
        if attr["type"] == "number":
            Player["PlayerAttributes"][attr["name"]] = {
                "N": random.randint(0, 100)}
    gamelift.start_matchmaking(ConfigurationName=CONFIG_NAME, Players=[Player])
    add_matching_players_list(Player)


def get_matchmaking_result():
    sqs = boto3.resource("sqs")
    queue = sqs.get_queue_by_name(QueueName="MatchResultQueue")

    for message in queue.receive_messages(MaxNumberOfMessages=10):
        match = json.loads(message.body)["detail"]
        integrate_results_players(
            match["gameSessionInfo"]["players"], match["matchId"])
        message.delete()


# Enqueue players details from Matchmaking results and put CloudWatch metric data point
@retry(stop_max_attempt_number=3)
def integrate_results_players(players, matchId):
    match = {"matchId": matchId, "players": {}}
    for player in players:
        pid = player["playerId"]
        if pid in _waiting_players.keys():
            match["players"][pid] = _waiting_players.pop(pid)
            match["players"][pid]["Elapsed"] = {
                "N": elapsed_time() - match["players"][pid]["StartTime"]["N"]
            }
            del match["players"][pid]["name"]
            del match["players"][pid]["StartTime"]

    enqueue_matched_players(match)
    put_metrics(match)


@retry(stop_max_attempt_number=3)
def put_metric_data(name, value, unit):
    cloudwatch = boto3.client("cloudwatch")
    cloudwatch.put_metric_data(
        Namespace="MatchMakingResults",
        MetricData=[
            {
                "MetricName": name,
                "Dimensions": [
                    {"Name": "RuleSetName", "Value": _rule_set_name},
                    {"Name": "MatchMakingConfigName", "Value": CONFIG_NAME},
                ],
                "Timestamp": datetime.datetime.now(),
                "Value": value,
                "Unit": unit,
            }
        ]
    )


# Retrieve player info from temporary list and put Metrics
def put_metrics(match):
    player_len = len(match["players"])
    if player_len < 1:
        return None

    attrs_min = {}
    attrs_max = {}
    for v in match["players"].values():
        for vk, vv in v.items():
            if "N" not in vv.keys():
                continue
            if not vk in attrs_min:
                attrs_min.setdefault(vk, vv["N"])
                attrs_max.setdefault(vk, vv["N"])
            attrs_min[vk] = min(vv["N"], attrs_min[vk])
            attrs_max[vk] = max(vv["N"], attrs_max[vk])

    avg_elapsed = (
        sum([v["Elapsed"]["N"]
            for v in match["players"].values()]) / player_len
    )
    put_metric_data("PlayersInMatch", player_len, "Count")
    put_metric_data("TotalTimeToMatch", avg_elapsed, "Seconds")

    for key in attrs_min.keys():
        put_metric_data(
            "Distance" + key +
            "InMatch", attrs_max[key] - attrs_min[key], "Count"
        )


# Display matchmaking status in terminal
def monitor_players_status(players_attrs_def):
    print("{}[2J{}[;H".format(chr(27), chr(27)))  # Clear Terminal Window
    attrs_names = [attr["name"] for attr in players_attrs_def]
    attrs_names.append("Start(s)")
    attrs_names.append("Name")
    attrs_names.append("   |   ")
    attrs_names.extend([attr["name"] for attr in players_attrs_def])
    attrs_names.append("Elapse(s)")
    attrs_names.append("Name")

    left_rows = []
    for pk, pv in _waiting_players.items():
        pv["name"] = {"S": pk}
        left_rows.append(
            "".join(
                ["{:10s}".format(str(next(iter(item.values()))))
                 for item in pv.values()]
            )
        )

    right_rows = []
    for item in list(_matched_sessions_queue):
        right_rows.append("=== matched pair ===  : {}".format(item["matchId"]))

        for pk, pv in item["players"].items():
            pv["name"] = {"S": pk}
            right_rows.append(
                "".join(
                    [
                        "{:10s}".format(str(next(iter(item.values()))))
                        for item in pv.values()
                    ]
                )
            )
        right_rows.append("")
        

    print("Elapsed time ...{}(s)".format(int(time.time() - _start_time)))
    print("".join(["{:10s}".format(str(item)) for item in attrs_names]))
    print("{}".format("_" * 72))

    for l, r in zip_longest(
        left_rows, right_rows, fillvalue="{:10s}".format("") * 3
    ):
        print("{}{:10s}{}".format(l,"   |   ", r))


def main():
    players_attrs_def = get_player_attr_definition()
    with ThreadPoolExecutor(10) as executor:
        while True:
            for i in range(4):
                executor.submit(start_matchmaking(players_attrs_def))
            executor.submit(get_matchmaking_result())
            executor.submit(monitor_players_status(players_attrs_def))
            time.sleep(REQUEST_INTERVAL)
            clear_matched_queue()


if __name__ == "__main__":
    main()
